==> ns-1.0a17.tar.gz: ns-1.0a17/node.cc <==
/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: node.cc,v 1.11 95/12/13 18:20:42 floyd Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>

#include "packet.h"
#include "node.h"
#include "link.h"
#include "agent.h"

static class NodeMatcher : public Matcher {
public:
	NodeMatcher() : Matcher("node") {}
	TclObject* match(const char* id) {
		return (new Node());
	}
} matcher_node;

int Node::cnt_ = 0;

Node::Node() : nport_(0), route_(0), links_(0), agentLock_(0)
{
	addr_ = cnt_++;
	maxport_ = 64;
	demux_ = new Agent*[maxport_];
	memset(mdemux_, 0, sizeof(mdemux_));
}

Node::~Node()
{
	delete route_;
	delete demux_;
	int i;
	for (i = GROUP_HASHSIZE; --i >= 0; )
		free(mdemux_[i]);
}

void Node::free(AgentList* p)
{
	while (p != 0) {
		AgentList* n = p->next;
		delete p;
		p = n;
	}
}

void Node::add_link(Link* p)
{
	p->src(addr_);
	p->next_ = (links_ != 0) ? links_->link : 0;
	LinkState* ls = new LinkState;
	memset(ls, 0, sizeof(*ls));
	ls->link = p;
	ls->next = links_;
	links_ = ls;
}

void Node::setroute(int dst, Link* p)
{
	if (route_ == 0) {
		route_ = new Link*[cnt_];
		memset(route_, 0, cnt_* sizeof(Link*));
	}
	route_[dst] = p;
}

int Node::add_agent(Agent* p)
{
	int port = nport_++;
	if (port > maxport_) {
		int n = maxport_ * 2;
		Agent** demux = new Agent*[n];
		for (int i = 0; i < maxport_; ++i)
			demux[i] = demux_[i];
		delete demux_;
		demux_ = demux;
		maxport_ = n;
	}
	demux_[port] = p;
	return (port);
}

void Node::join_group(Agent* agent, int group)
{
	/*
	 * If group has been previously pruned,
	 * send a graft.
	 */
	send_grafts(group);

	int h = GROUP_HASH(group);
	AgentList* al = new AgentList;
	al->group = group;
	al->agent = agent;
	al->deleted = 0;
	al->next = mdemux_[h];
	mdemux_[h] = al;
}

void Node::leave_group(Agent* agent, int group)
{
	int h = GROUP_HASH(group);
	AgentList** p;
	for (p = &mdemux_[h]; ; p = &(*p)->next) {
		AgentList* target = *p;
		if (target->agent == agent && target->group == group) {
			if (agentLock_) {
				target->deleted = 1;
				agentLock_ = -1;
			} else {
				*p = (*p)->next;
				delete target;
			}
			return;
		}
	}
}

void Node::delete_agents(int group)
{
	int h = GROUP_HASH(group);
	AgentList** p = &mdemux_[h];
	while (*p != 0) {
		if ((*p)->deleted) {
			AgentList* target = *p;
			*p = (*p)->next;
			delete target;
		} else
			p = &(*p)->next;
	}
}

void Node::send_graft(int group, Link* link)
{
	Packet* pkt = Packet::alloc();
	memset(pkt, 0, sizeof(*pkt));
	pkt->type_ = PT_GRAFT;
	/*XXX*/
	pkt->src_ = addr_;
	pkt->dst_ = group;
	pkt->class_ = 8;/*XXX*/
	pkt->size_ = 40;/*XXX*/
	link->send(pkt);
}

/*
 * Send graft packets along the RPF tree for each source that
 * we've seen send packets to the specified group.  Send only
 * one graft per link.
 */
void Node::send_grafts(int group)
{
	for (LinkState* ls = links_; ls != 0; ls = ls->next) {
		if (member(ls->prune_up, group)) {
			delete_member(ls->prune_up, group);
			/*XXX should be reliable */
			send_graft(group, ls->link);
		}
	}
}

void Node::record_prune(const Packet* p)
{
	Link* link = route_[p->lsrc_];
	int group = p->dst_;
	for (LinkState* ls = links_; ls != 0; ls = ls->next) {
		if (ls->link == link) {
			if (!member(ls->prune_up, group))
				add_member(ls->prune_up, group);
			return;
		}
	}
	abort();
}

void Node::send_prune(Packet* p)
{
	/*
	 * XXX assume route gives link that packet arrived on
	 * (otherwise, there are multiple links or routing algorithm
	 * has a bug.)
	 */
	Link* link = route_[p->lsrc_];
	p->type_ = PT_PRUNE;
	p->size_ = 40;/*XXX configurable*/
	p->class_ |= 32; /*XXX*/
	link->send(p);
}

int Node::member(groupEnt** hashtab, int g)
{
	for (groupEnt* p = hashtab[GROUP_HASH(g)]; p != 0; p = p->next)
		if (p->group == g)
			return (1);
	return (0);
}

void Node::add_member(groupEnt** hashtab, int g)
{
	int h = GROUP_HASH(g);
	groupEnt* p = new groupEnt;
	p->group = g;
	p->next = hashtab[h];
	hashtab[h] = p;
}

void Node::delete_member(groupEnt** hashtab, int g)
{
	int h = GROUP_HASH(g);
	groupEnt** p;
	for (p = &hashtab[h]; (*p)->group != g; p = &(*p)->next)
		;
	groupEnt* target = *p;
	*p = (*p)->next;
	delete target;
}

void Node::multicast_forward(Packet* p)
{
	int g = p->dst_;
	if (p->type_ == PT_PRUNE) {
		/*
		 * A prune message.  Add the group in question
		 * to the list of pruned addresses for the
		 * interface that the prune message arrived on.
		 * XXX and set a timer to time it out.
		 */
		for (LinkState* ls = links_; ls != 0; ls = ls->next) {
			if (ls->link->dst() == p->lsrc_) {
				if (!member(ls->prune_down, g))
					add_member(ls->prune_down, g);
				Packet::free(p);
				return;
			}
		}
		/* XXX should have found the link */
		abort();
	} else if (p->type_ == PT_GRAFT) {
		/*
		 * A graft message.  Propagate the graft to
		 * all any upstream flows that have been pruned.
		 */
		send_grafts(p->dst_);
		/*
		 * Delete the group in question from the prune list
		 * on the interface that the graft message arrived on.
		 */
		for (LinkState* ls = links_; ls != 0; ls = ls->next) {
			if (ls->link->dst() == p->lsrc_) {
				if (member(ls->prune_down, g))
					delete_member(ls->prune_down, g);
				Packet::free(p);
				return;
			}
		}
		abort();
	}
	/*
	 * Check if packet arrive on wrong interface.
	 * If so, send a prune message back up the RPF tree.
	 * (This is a natural consequence of the dense-mode
	 * PIM-style flooding.)
	 */
	if (p->src_ != addr_ && route_[p->src_]->dst() != p->lsrc_) {
		send_prune(p);
		return;
	}

	/*
	 * This is a multicast packet.  Forward it to all
	 * nodes along the reverse-path forwarding tree,
	 * and deliver to all local agents that are members
	 * of the corresponding multicast group.
	 */

	int copies = 0;
	Link* reverse = route_[p->src_];
	for (LinkState* ls = links_; ls != 0; ls = ls->next) {
		/*
		 * Forward the packet along each link that
		 * isn't pruned and isn't the inbound edge
		 * of the RPF tree.
		 */
		Link* link = ls->link;
		if (link == reverse)
			continue;
		if (!member(ls->prune_down, g)) {
			link->send(p->copy());
			++copies;
		}
	}
	/*
	 * Demux multicast packets to any local agents.
	 * (Note that we don't use the port for mcast demux.)
	 * Be careful to lock the agentlist data structure
	 * so that if an upcall into tcl is made and tcl in turn
	 * does a leave-group, we don't end up with a pointer to
	 * garbage.
	 */
	agentLock_ = 1;
	AgentList* local = mdemux_[GROUP_HASH(g)];
	for (; local != 0; local = local->next) {
		if (local->group != g || local->deleted)
			continue;
		local->agent->recv(p->copy());
		++copies;
	}
	if (agentLock_ < 0)
		delete_agents(g);
	agentLock_ = 0;

	if (copies == 0 && p->src_ != addr_) {
		/*
		 * If traffic flow is not needed at this node,
		 * the packet source isn't on this node,
		 * prune flow back.
		 */
		record_prune(p);
		send_prune(p);
	} else
		Packet::free(p);
}

void Node::unicast_forward(Packet* p)
{
	if (p->dst_ == addr_) {
		/* packet destined to this node, lookup agent */
		int port = p->dport_;
		if (port > nport_)
			abort();
		demux_[port]->recv(p);
		return;
	}
	Link* link = route_[p->dst_];
	if (link == 0) {
		/*XXX drop packet and bump error counter */
		fprintf(stderr, "no route for packet\n");
		exit(1);
	}
	link->send(p);
}

/*
 * Nodes only ever see "packet" events, which come either
 * from an incoming link or a local agent (i.e., packet source).
 */
void Node::handle(Event* e)
{
	if (route_ == 0) {
		fprintf(stderr, "ns: can't run without routes\n");
		exit(1);
	}
	Packet* p = (Packet*)e;
	if (p->dst_ & MCASTADDR)
		multicast_forward(p);
	else
		unicast_forward(p);
}

/*
 * $node addr
 * $node agent $port
 */
int Node::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "addr") == 0) {
			tcl.resultf("%d", addr_);
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "agent") == 0) {
			int port = atoi(argv[2]);
			if (port < maxport_) {
				Agent* a = demux_[port];
				if (a != 0)
					tcl.result(a->TclObject::name());
			}
			return (TCL_OK);
		}
	}
	return (NsObject::command(argc, argv));
}

==> ns-1.1.tar.gz: ns-1.1/node.cc <==
/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: node.cc,v 1.15 96/02/19 14:34:25 mccanne Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>

#include "packet.h"
#include "node.h"
#include "link.h"
#include "agent.h"

static class NodeMatcher : public Matcher {
public:
	NodeMatcher() : Matcher("node") {}
	TclObject* match(const char* id) {
		return (new Node());
	}
} matcher_node;

int Node::cnt_ = 0;

Node::Node() : nport_(0), route_(0), links_(0), agentLock_(0)
{
	addr_ = cnt_++;
	maxport_ = 64;
	demux_ = new Agent*[maxport_];
	for (int i = 0; i < maxport_; ++i)
		demux_[i] = 0;
	memset(mdemux_, 0, sizeof(mdemux_));
}

Node::~Node()
{
	delete route_;
	delete demux_;
	int i;
	for (i = GROUP_HASHSIZE; --i >= 0; )
		free(mdemux_[i]);
}

void Node::free(AgentList* p)
{
	while (p != 0) {
		AgentList* n = p->next;
		delete p;
		p = n;
	}
}

void Node::add_link(Link* p)
{
	p->src(addr_);
	p->next_ = (links_ != 0) ? links_->link : 0;
	LinkState* ls = new LinkState;
	memset(ls, 0, sizeof(*ls));
	ls->link = p;
	ls->next = links_;
	links_ = ls;
}

void Node::setroute(int dst, Link* p)
{
	if (route_ == 0) {
		route_ = new Link*[cnt_];
		memset(route_, 0, cnt_* sizeof(Link*));
	}
	route_[dst] = p;
}

int Node::add_agent(Agent* p)
{
	int port = nport_++;
	if (port >= maxport_) {
		int n = maxport_ * 2;
		Agent** demux = new Agent*[n];
		int i;
		for (i = 0; i < maxport_; ++i)
			demux[i] = demux_[i];
		for (; i < n; ++i)
			demux[i] = 0;
		delete demux_;
		demux_ = demux;
		maxport_ = n;
	}
	demux_[port] = p;
	return (port);
}

void Node::join_group(Agent* agent, int group)
{
	/*
	 * If group has been previously pruned,
	 * send a graft.
	 */
	(void)send_grafts(group);

	int h = GROUP_HASH(group);
	AgentList* al = new AgentList;
	al->group = group;
	al->agent = agent;
	al->deleted = 0;
	al->next = mdemux_[h];
	mdemux_[h] = al;
}

void Node::leave_group(Agent* agent, int group)
{
	int h = GROUP_HASH(group);
	AgentList** p;
	for (p = &mdemux_[h]; ; p = &(*p)->next) {
		AgentList* target = *p;
		if (target->agent == agent && target->group == group) {
			if (agentLock_) {
				target->deleted = 1;
				agentLock_ = -1;
			} else {
				*p = (*p)->next;
				delete target;
			}
			return;
		}
	}
}

void Node::delete_agents(int group)
{
	int h = GROUP_HASH(group);
	AgentList** p = &mdemux_[h];
	while (*p != 0) {
		if ((*p)->deleted) {
			AgentList* target = *p;
			*p = (*p)->next;
			delete target;
		} else
			p = &(*p)->next;
	}
}

void Node::send_graft(int group, Link* link)
{
	Packet* pkt = Packet::alloc();
	memset(pkt, 0, sizeof(*pkt));
	pkt->type_ = PT_GRAFT;
	/*XXX*/
	pkt->src_ = addr_;
	pkt->dst_ = group;
	pkt->class_ = 8;/*XXX*/
	pkt->size_ = 40;/*XXX*/
	link->send(pkt);
}

/*
 * Send graft packets along the RPF tree for each source that
 * we've seen send packets to the specified group.  Send only
 * one graft per link.  Note: prune_up keeps track of the
 * groups for which we've sent prunes up the tree, while prune_down
 * keeps track of groups for which we are not forwarding down the tree.
 *
 * Returns true iff no grafts needed to be propagated.
 */
int Node::send_grafts(int group)
{
	int none_sent = 1;
	for (LinkState* ls = links_; ls != 0; ls = ls->next) {
		if (member(ls->prune_up, group)) {
			delete_member(ls->prune_up, group);
			/*XXX should be reliable */
			send_graft(group, ls->link);
			none_sent = 0;
		}
	}
	return (none_sent);
}

void Node::record_prune(const Packet* p)
{
	Link* link = route_[p->lsrc_];
	int group = p->dst_;
	for (LinkState* ls = links_; ls != 0; ls = ls->next) {
		if (ls->link == link) {
			if (!member(ls->prune_up, group))
				add_member(ls->prune_up, group);
			return;
		}
	}
	abort();
}

void Node::send_prune(Packet* p)
{
	/*
	 * XXX assume route gives link that packet arrived on
	 * (otherwise, there are multiple links or routing algorithm
	 * has a bug.)
	 */
	Link* link = route_[p->lsrc_];
	if (link == 0)
		abort();
	p->type_ = PT_PRUNE;
	p->size_ = 40;/*XXX configurable*/
	p->class_ |= 32; /*XXX*/
	link->send(p);
}

int Node::member(groupEnt** hashtab, int g)
{
	for (groupEnt* p = hashtab[GROUP_HASH(g)]; p != 0; p = p->next)
		if (p->group == g)
			return (1);
	return (0);
}

void Node::add_member(groupEnt** hashtab, int g)
{
	int h = GROUP_HASH(g);
	groupEnt* p = new groupEnt;
	p->group = g;
	p->next = hashtab[h];
	hashtab[h] = p;
}

void Node::delete_member(groupEnt** hashtab, int g)
{
	int h = GROUP_HASH(g);
	groupEnt** p;
	for (p = &hashtab[h]; (*p)->group != g; p = &(*p)->next)
		;
	groupEnt* target = *p;
	*p = (*p)->next;
	delete target;
}

/*
 * Send an explicit-join control message back down the
 * RPF tree defined by src (the packet will look like
 * it came from src).  Send it to "hop", which is the
 * the node that sent us the graft that prompted the
 * explicit-join.
 *
 * This experimental code related to RLM.
 */
void Node::send_join_control(int src, int g, int hop)
{
	Packet* pkt = Packet::alloc();
	memset(pkt, 0, sizeof(*pkt));
	pkt->type_ = PT_MESSAGE;
	/*XXX*/
	pkt->src_ = src;
	pkt->dst_ = (g & ~0xf) + 0x1000;
	pkt->class_ = 8;/*XXX*/
	pkt->size_ = 40;/*XXX*/
	/* tell downstream cloud the layer being added */
	sprintf(pkt->bd_.msg_, "x-join %d", g & 0xf);
	route_[hop]->send(pkt);
}

/*XXX hack: since we are not doing source-specific prunes, we need
 * a way to map a group back to the source (and we assume there
 * is only one source per group).
 */
int hack_group_to_source[1024];

void Node::multicast_forward(Packet* p)
{
	int g = p->dst_;

	if (p->type_ == PT_PRUNE) {
		/*
		 * A prune message.  Add the group in question
		 * to the list of pruned addresses for the
		 * interface that the prune message arrived on.
		 * XXX and set a timer to time it out.
		 */
		for (LinkState* ls = links_; ls != 0; ls = ls->next) {
			if (ls->link->dst() == p->lsrc_) {
				if (!member(ls->prune_down, g))
					add_member(ls->prune_down, g);
				Packet::free(p);
				return;
			}
		}
		/* XXX should have found the link */
		abort();
	} else if (p->type_ == PT_GRAFT) {
		/*
		 * A graft message.  Propagate the graft to
		 * all any upstream flows that have been pruned.
		 * Assume that MMGs are in the range 0x8000-0x8fff.
		 */
		if (send_grafts(g) && g >= 0x8000 && g < 0x9000) {
			/*
			 * no grafts needed to be sent up the RPF tree.
			 * multicast a local join message back down
			 * the tree on the control address.
			 * XXX should be configurable
			 */
			int src = hack_group_to_source[g & 1023];
			send_join_control(src, g, p->lsrc_);
		}
		/*
		 * Delete the group in question from the prune list
		 * on the interface that the graft message arrived on.
		 */
		for (LinkState* ls = links_; ls != 0; ls = ls->next) {
			if (ls->link->dst() == p->lsrc_) {
				if (member(ls->prune_down, g))
					delete_member(ls->prune_down, g);
				Packet::free(p);
				return;
			}
		}
		abort();
	}
	hack_group_to_source[g & 1023] = p->src_;

	/*
	 * Check if packet arrive on wrong interface.
	 * If so, send a prune message back up the RPF tree.
	 * (This is a natural consequence of the dense-mode
	 * PIM-style flooding.)
	 */
	if (p->src_ != addr_ && route_[p->src_]->dst() != p->lsrc_) {
		send_prune(p);
		return;
	}

	/*
	 * This is a multicast packet.  Forward it to all
	 * nodes along the reverse-path forwarding tree,
	 * and deliver to all local agents that are members
	 * of the corresponding multicast group.
	 */

	int copies = 0;
	Link* reverse = route_[p->src_];
	for (LinkState* ls = links_; ls != 0; ls = ls->next) {
		/*
		 * Forward the packet along each link that
		 * isn't pruned and isn't the inbound edge
		 * of the RPF tree.
		 */
		Link* link = ls->link;
		if (link == reverse)
			continue;
		if (!member(ls->prune_down, g)) {
			link->send(p->copy());
			++copies;
		}
	}
	/*
	 * Demux multicast packets to any local agents.
	 * (Note that we don't use the port for mcast demux.)
	 * Be careful to lock the agentlist data structure
	 * so that if an upcall into tcl is made and tcl in turn
	 * does a leave-group, we don't end up with a pointer to
	 * garbage.
	 */
	agentLock_ = 1;
	AgentList* local = mdemux_[GROUP_HASH(g)];
	for (; local != 0; local = local->next) {
		if (local->group != g || local->deleted)
			continue;
		local->agent->recv(p->copy());
		++copies;
	}
	if (agentLock_ < 0)
		delete_agents(g);
	agentLock_ = 0;

	if (copies == 0 && p->src_ != addr_) {
		/*
		 * If traffic flow is not needed at this node,
		 * the packet source isn't on this node,
		 * prune flow back.
		 */
		record_prune(p);
		send_prune(p);
	} else
		Packet::free(p);
}

void Node::unicast_forward(Packet* p)
{
	if (p->dst_ == addr_) {
		/* packet destined to this node, lookup agent */
		int port = p->dport_;
		if (port > nport_)
			abort();
		Agent* a = demux_[port];
		if (a != 0)
			a->recv(p);
		else
			Packet::free(p);
		return;
	}
	Link* link = route_[p->dst_];
	if (link == 0) {
		/*XXX drop packet and bump error counter */
		fprintf(stderr, "no route for packet\n");
		exit(1);
	}
	link->send(p);
}

/*
 * Nodes only ever see "packet" events, which come either
 * from an incoming link or a local agent (i.e., packet source).
 */
void Node::handle(Event* e)
{
	if (route_ == 0) {
		fprintf(stderr, "ns: can't run without routes\n");
		exit(1);
	}
	Packet* p = (Packet*)e;
	if (p->dst_ & MCASTADDR)
		multicast_forward(p);
	else
		unicast_forward(p);
}

/*
 * $node addr
 * $node agent $port
 */
int Node::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "addr") == 0) {
			tcl.resultf("%d", addr_);
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "agent") == 0) {
			int port = atoi(argv[2]);
			if (port < maxport_) {
				Agent* a = demux_[port];
				if (a != 0)
					tcl.result(a->TclObject::name());
			}
			return (TCL_OK);
		}
	}
	return (NsObject::command(argc, argv));
}
